/*
 * Obliterate - a simple program to obliterate file contents.
 *
 * Copyright (c) 1999 Softweyr LLC, South Jordan, Utah USA.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY Softweyr LLC ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL Softweyr LLC OR ANY CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 * 
 * Author: Wes Peters
 * Date: Mon Jul 12 15:04:51 MDT 1999
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>

#include <stdio.h>
#include <fcntl.h>

#define MIN(a,b) ((a < b) ? a : b)

#define DEVICE_BUFFER_SIZE 8192

/*
 * Option flags.
 */
static int force = 0;
static int verbose = 0;

#define vprintf if (verbose) printf
#define vvprintf if (verbose > 1) fflush(stdout), printf

char *progname;

/*
 * Status values
 */
typedef enum { success, fail } status_t;

/*
 * The bit patterns used to overwrite the disk blocks.  These patterns are
 * designed to provide a complete overwrite on all common disk recording
 * formats.
 *
 * Pattern devised by Peter Gutmann, Department of Computer Science,
 * University of Auckland <pgut001@cs.auckland.ac.nz>.  See Peter's paper
 * "Secure Deletion of Data from Magnetic and Solid-State Memory" at
 * http://www.cs.auckland.ac.nz/~pgut001/pubs/secure_del.html for more
 * information about how and why this pattern works.
 */

unsigned char *patterns[] =
{
    "R",		/* random */
    "R",		/* random */
    "R",		/* random */
    "R",		/* random */
    "\x55",		/* 1,7 RLL         MFM */
    "\xaa",		/* 1,7 RLL         MFM */
    "\x92\x49\x24",	/*         2,7 RLL MFM */
    "\x49\x24\x92",	/*         2,7 RLL MFM */
    "\x24\x92\x49",	/*         2,7 RLL MFM */
    "\x00",		/* 1,7 RLL 2,7 RLL     */
    "\x11",		/* 1,7 RLL             */
    "\x22",		/* 1,7 RLL             */
    "\x33",		/* 1,7 RLL 2,7 RLL     */
    "\x44",		/* 1,7 RLL             */
    "\x55",		/* 1,7 RLL         MFM */
    "\x66",		/* 1,7 RLL 2,7 RLL     */
    "\x77",		/* 1,7 RLL             */
    "\x88",		/* 1,7 RLL             */
    "\x99",		/* 1,7 RLL 2,7 RLL     */
    "\xaa",		/* 1,7 RLL         MFM */
    "\xbb",		/* 1,7 RLL             */
    "\xcc",		/* 1,7 RLL 2,7 RLL     */
    "\xdd",		/* 1,7 RLL             */
    "\xee",		/* 1,7 RLL             */
    "\xff",		/* 1,7 RLL 2,7 RLL     */
    "\x92\x49\x24",	/*         2,7 RLL MFM */
    "\x49\x24\x92",	/*         2,7 RLL MFM */
    "\x24\x92\x49",	/*         2,7 RLL MFM */
    "\x6d\xb6\xdb",	/*         2,7 RLL     */
    "\xb6\xdb\x6d",	/*         2,7 RLL     */
    "\xdb\x6d\xb6",	/*         2,7 RLL     */
    "R",		/* random */
    "R",		/* random */
    "R",		/* random */
    "R",		/* random */
};

static int npatterns = sizeof (patterns) / sizeof (patterns[0]);

/*
 * A file handle to access the system pool of randomness.
 */

int randfd = 0;

/*
 * Fill a block of memory with a specified pattern, repeating the pattern as
 * necessary to completely fill the block.  This routine is smart about
 * "special" patterns such as an apparent NULL pattern (this means zero-fill
 * the block) and "R" which means fill the block with random values.
 */

status_t
fill(void *data, off_t size, char *pattern)
{
    /*
     * If the first byte is zero, it means zero-fill the block.
     */
    if (*pattern == '\0')
    {
	(void) memset(data, 0, size);
    }
    /*
     * If the first byte is 'R' it means random fill the block.  We use the
     * system pool of randomness here because it is an easy to use source of
     * pretty much random data.  Reading blocks of any appreciable size from
     * the pool probably degrades it into a cascade of MD5 signatures, but
     * should be sufficient for our needs here.
     */
    else if (*pattern == 'R')
    {
	read(randfd, data, size);
    }
    else
    {
	/*
	 * Be smart about using library functions where available.
	 */
	int l = strlen(pattern);
	if (l == 1)
	{
	    memset(data, *pattern, size);
	}
	else
	{
	    char *p, *e = data + size;
	    for (p = data; p < e; p += l)
	    {
		memcpy(p, pattern, MIN(l, e - p));
	    }
	}
    }
}


/*
 * Overwrite a device.  We can't map a device into memory, and so have to
 * resort to write(2) et al.  Return 0 on success, -1 otherwise.
 */
status_t
obliterate_device(char *fname)
{
    char buffer[DEVICE_BUFFER_SIZE];
    int iteration, fd, count;
    ssize_t n;
    
    /*
     * Overwrite the file with successive patterns as specified above.  This
     * uses a buffer and multiple writes to overwrite the device.  Since we
     * only have (non-seekable) character devices, we can't even rewind,
     * just re-open the file each time.  Ugh.
     */
    vvprintf("\n");
    for (iteration = 0; iteration < npatterns; iteration++)
    {
	count = 0;
	fd = open(fname, O_RDWR);
	if (fd == -1)
	{
	    vprintf("\n");
	    perror(fname);
	    return fail;
	}
	
	fill(buffer, DEVICE_BUFFER_SIZE, patterns[iteration]);
	while ((n = write(fd, buffer, DEVICE_BUFFER_SIZE)) == DEVICE_BUFFER_SIZE)
	{
	    count++;
	}

	close(fd);
	vvprintf("\rpass %2d of %2d (%d + %d of %02x) ",
		 iteration + 1, npatterns, count, n, *patterns[iteration]);
    }

    vprintf("obliterated ");

    return success;
}


/*
 * Overwrite a file.  Return 0 on success, -1 otherwise.
 */
status_t
obliterate_file(char *fname, struct stat S)
{
    int fd;
    void *data;
    int iteration;
    
    fd = open(fname, O_RDWR);
    if (fd == -1)
    {
	if (force)
	{
	    /*
	     * It may have failed due to perimissions.  Try to modify it so we
	     * can write it then retry the open.
	     */
	    if (chmod(fname, 0600) == -1)
	    {
		vprintf("\n");
		perror(fname);
		return fail;
	    }
	    
	    fd = open(fname, O_RDWR);
	    if (fd == -1)
	    {
		vprintf("\n");
		perror(fname);
		return fail;
	    }
	}
	else
	{
	    vprintf("\n");
	    perror(fname);
	    return fail;
	}
    }

    vprintf("opened ");
    
    data = mmap(NULL, S.st_size, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
    if (data == MAP_FAILED)
    {
	vprintf("\n");
	perror(fname);
	close(fd);
	return fail;
    }

    vprintf("mapped ");

    /*
     * Overwrite the file with successive patterns as specified above.
     */
    for (iteration = 0; iteration < npatterns; iteration++)
    {
	fill(data, S.st_size, patterns[iteration]);
	if (msync(data, S.st_size, MS_SYNC) == -1)
	{
	    vprintf("\n");
	    perror(fname);
	    close(fd);
	    return fail;
	}

	vvprintf("\n pass %2d of %2d ", iteration + 1, npatterns);
    }
    
    vprintf("obliterated ");

    /*
     * Get rid of the file (from the filesystem point of view).
     */
    if (munmap(data, S.st_size) == -1)
    {
	vprintf("\n");
	perror(fname);
	close(fd);
	return fail;
    }

    vprintf("unmapped ");

    if (fsync(fd) == -1)
    {
	vprintf("\n");
	perror(fname);
	close(fd);
	return fail;
    }

    if (close(fd) == -1)
    {
	vprintf("\n");
	perror(fname);
	close(fd);
	return fail;
    }

    if (unlink(fname) == -1)
    {
	vprintf("\n");
	perror(fname);
	return fail;
    }

    vprintf("removed.\n");

    return success;
}


/*
 * Overwrite the named filesystem object.  Return 0 on success, -1 otherwise.
 */

status_t
obliterate(char *fname)
{
    struct stat S;
    status_t status = fail;

    vprintf("obliterating %s: ", fname);

    if (stat(fname, &S) == -1)
    {
	vprintf("\n");
	perror(fname);
	return fail;
    }

    if (S_ISCHR(S.st_mode))
    {
	if (force) return obliterate_device(fname);

	vprintf("\n");
	fprintf(stderr, "%s: not a regular file, specify -f to force\n", fname);
	return fail;
    }
    else if (S_ISREG(S.st_mode))
    {
	return obliterate_file(fname, S);
    }

    vprintf("\n");
    fprintf(stderr, "%s: not a regular file or disk partition\n", fname);
    return fail;
}


void
usage()
{
    fprintf(stderr,
	    "usage: %s [ -fv ] file [ files...]\n"
	    "\t-f: forcibly %s file, even if read-only\n"
	    "\t-v: be verbose\n",
	    progname, progname);
    exit(1);
}


/*
 * Obliterate each file named on the command line.
 */

int
main(int argc, char *argv[])
{
    status_t result = success;
    extern char *optarg;
    extern int optind;
    int ch, fd;

    progname = argv[0];
    
    while ((ch = getopt(argc, argv, "fv")) != -1)
    {
	switch (ch)
	{
	case 'f':
	    force++;
	    break;
	case 'v':
	    verbose++;
	    break;
	case '?':
	default:
	    usage();
	}
    }
    
    argc -= optind;
    argv += optind;

    if (argc < 1)
    {
	usage();
    }
    
    /*
     * Open the system entropy pool for random data.
     */
    if ((randfd = open("/dev/urandom", O_RDONLY)) == -1)
    {
	perror("Opening system pool of randomness (/dev/urandom)");
	return fail;
    }
    
    while (argc--)
    {
	result |= obliterate(*argv++);
    }
    
    close(randfd);
    return result;
}
